#!/usr/bin/env python3
"""
Assembly Operations in Python
by Javantea
Feb 4, 2015

Describe the event of an instruction to Python.
For example:
python_object = AsmNoop() allows a person to query what happens to EBX? 
We can say that EBX is ebx.
python_object.stepEBX(44) == 44

"""

class AsmOperation:
	def __init__(self):
		pass
	#end def AsmOperation()
	
	def stepEAX(self, eax):
		return eax
	#end def stepEAX(eax)
	
	def stepEBX(self, ebx):
		return ebx
	#end def stepEBX(ebx)
	
	def stepECX(self, ecx):
		return ecx
	#end def stepECX(ecx)
	
	def stepEDX(self, edx):
		return edx
	#end def stepEDX(edx)
	
	def stepEDI(self, edi):
		return edi
	#end def stepEDI(edi)
	
	def stepESI(self, esi):
		return esi
	#end def stepESI(esi)
	
	def stepESP(self, esp):
		return esp
	#end def stepESP(esp)
	
	def stepEBP(self, ebp):
		return ebp
	#end def stepEBP(ebp)
	
	def stepEIP(self, eip, length):
		return eip + length
	#end def stepEIP(eip, length)
#end class AsmOperation

class AsmNoop(AsmOperation):
	pass
#end class AsmNoop(AsmOperation)

class AsmAddImmReg(AsmOperation):
	"""
	431: ADD EBX, 0x1bcf; EBX += 0x1bcf;
	"""
	def __init__(self, reg, imm):
		self.reg = reg
		self.imm = imm
	#end def AsmAddImmReg(reg, imm)
	
	def stepAny(self, reg, value):
		if reg == self.reg:
			return value + self.imm
		#end if
		return value
	#end def stepAny(reg, value)
	
	def stepEAX(self, eax):
		return self.stepAny('EAX', eax)
	#end def stepEAX(eax)
	
	def stepEBX(self, eax):
		return self.stepAny('EBX', eax)
	#end def stepEAX(eax)
	
	def stepECX(self, ecx):
		return self.stepAny('ECX', ecx)
	#end def stepECX(ecx)
	
	def stepEDX(self, edx):
		return self.stepAny('EDX', edx)
	#end def stepEDX(edx)
	
	def stepEDI(self, edi):
		return self.stepAny('EDI', edi)
	#end def stepEDI(edi)
	
	def stepESI(self, esi):
		return self.stepAny('ESI', esi)
	#end def stepESI(esi)
	
	def stepESP(self, esp):
		return self.stepAny('ESP', esp)
	#end def stepESP(esp)
	
	def stepEBP(self, ebp):
		return self.stepAny('EBP', ebp)
	#end def stepEBP(ebp)
#end class AsmAddImmReg(AsmOperation)

class AsmSubImmReg(AsmAddImmReg):
	"""
	429: SUB ESP, 0x8; ESP -= 0x8;
	"""
	def __init__(self, reg, imm):
		self.reg = reg
		self.imm = imm
	#end def AsmAddImmReg(reg, imm)
	
	def stepAny(self, reg, value):
		if reg == self.reg:
			return value - self.imm
		#end if
		return value
	#end def stepAny(reg, value)
#end class AsmSubImmReg(AsmOperation)

class AsmMulImmReg(AsmAddImmReg):
	"""
	40feba: IMUL EDX, EDX, 0x190
	"""
	def __init__(self, reg, imm):
		self.reg = reg
		self.imm = imm
	#end def AsmAddImmReg(reg, imm)
	
	def stepAny(self, reg, value):
		if reg == self.reg:
			return value * self.imm
		#end if
		return value
	#end def stepAny(reg, value)
#end class AsmMulImmReg(AsmOperation)

class AsmMovImmReg(AsmAddImmReg):
	"""
	402761: MOV EDX, 0xee2
	"""
	def __init__(self, reg, imm):
		self.reg = reg
		self.imm = imm
	#end def AsmAddImmReg(reg, imm)
	
	def stepAny(self, reg, value):
		if reg == self.reg:
			return self.imm
		#end if
		return value
	#end def stepAny(reg, value)
#end class AsmMovImmReg(AsmOperation)

def main():
	"""
	Validate the functionality with a simple asm program like this:
	MOV EAX, 0x1
	MOV EBX, 0xa
	SUB EAX, 0x8
	SUB EBX, 0xffff8
	ADD EAX, 0xff
	Then check the answers with Python assert.
	eax == (1-8+255)
	ebx == (10-0xffff8)
	"""
	a = AsmNoop()
	b = AsmSubImmReg('EAX', 0x8)
	c = AsmSubImmReg('EBX', 0xffff8)
	d = AsmAddImmReg('EAX', 0xff)
	eax = 1
	ebx = 10
	eax = a.stepEAX(eax)
	ebx = a.stepEBX(ebx)
	eax = b.stepEAX(eax)
	ebx = b.stepEBX(ebx)
	eax = c.stepEAX(eax)
	ebx = c.stepEBX(ebx)
	eax = d.stepEAX(eax)
	ebx = d.stepEBX(ebx)
	print(eax, ebx)
	assert(eax == (1-8+255))
	assert(ebx == (10-0xffff8))
#end def main()

if __name__ == '__main__':
	main()
#end if
