#!/usr/bin/env python
"""
Disassembler
by Javantea
Sept 27, 2014

TODO: Better 32-bit support, main function not found in if2_32
TODO: ARM support
TODO: Compare Python disassembly libraries
TODO: Consider using pyelftools
TODO: Python 3
TODO: Testing against junk binaries

A little easier to deal with than objcopy. It has most of the detail that 
objdump gives us and the flexibility of python. 

Note that this isn't Intel compatible or AT&T compatible. It uses Intel syntax 
for almost everything because it uses distorm. The CALL syntax is different 
than Intel, when it's a static call, it uses objdump's AT&T syntax. The reason 
for this isn't entirely obvious. To me, it's a little easier to read and parse 
despite it being a bit messy having a syntax that doesn't make sense.

The honest truth is that assemblers like NASM and YASM won't compile Intel 
syntax in the way that I print it out, so this is only useful for a programmer 
or a reverse engineer. However, you can use cut.

echo 'BITS 64' > abitofasm1.asm
cut -c 10- abitofasm.asm >> abitofasm1.asm
nasm -o abitofasm.o abitofasm1.asm

objdump -D -b binary -m i386 -M x86_64 abitofasm.o
abitofasm.o:     file format binary
Disassembly of section .data:
00000000 <.data>:
   0:   84 d2                   test   %dl,%dl
   2:   b9 01 00 00 00          mov    $0x1,%ecx
   7:   0f 85 3b 71 40 00       jne    0x407148
   d:   5b                      pop    %ebx
   e:   31 c9                   xor    %ecx,%ecx
  10:   5d                      pop    %ebp
  11:   89 c8                   mov    %ecx,%eax
  13:   41                      inc    %ecx
  14:   5c                      pop    %esp
  15:   c3                      ret    

As you can see the jump is absolute and not relative, so we have a problem that 
we need labels to solve.

Also, Intel syntax:
ndisasm -b 64 abitofasm.o
00000000  84D2              test dl,dl
00000002  B901000000        mov ecx,0x1
00000007  0F853B714000      jnz qword 0x407148
0000000D  5B                pop rbx
0000000E  31C9              xor ecx,ecx
00000010  5D                pop rbp
00000011  89C8              mov eax,ecx
00000013  415C              pop r12
00000015  C3                ret

FIXME:
  40a6cc: IMUL RAX, RAX, 0x3e8 ; RAX *= RAX,; 
TODO:
  40a759: MUL R9 ; 
  40a75f: LEA RAX, [RDX+RDX*2]
  40a789: DIV RCX ; 
  40a79f: MOV RAX, RDI

"""
from __future__ import print_function
import sys
try:
	import distorm3 as distorm
except ImportError:
	distorm = None
#end try
if distorm == None:
	try:
		import distorm
	except ImportError:
		sys.stderr.write("Error: Distorm is missing.\n")
		exit(1)
	#end try
#end if
import elf1
import struct
import string
import re
from asmop1 import AsmNoop, AsmAddImmReg, AsmSubImmReg, AsmMulImmReg, AsmMovImmReg
import os

def dl_new_hash(s):
	"""
	"Dan Bernstein's string hash function posted eons ago on comp.lang.c."
	https://sourceware.org/ml/binutils/2006-10/msg00377.html
	"""
	h = 5381
	for c in s:
		h = (h * 33 + ord(c)) & 0xffffffff
	#next c
	return h & 0xffffffff
#end def dl_new_hash(s)

"""
Directly from binutils-2.23.2/bfd/elf64-x86-64.c:
#define PLT_ENTRY_SIZE 16

#define GET_PLT_ENTRY_SIZE(abfd) \
  get_elf_x86_64_backend_data (abfd)->plt_entry_size

static bfd_vma
elf_x86_64_plt_sym_val (bfd_vma i, const asection *plt,
			const arelent *rel ATTRIBUTE_UNUSED)
{
  return plt->vma + (i + 1) * GET_PLT_ENTRY_SIZE (plt->owner);
}
"""
def elf_x86_64_plt_sym_val(i, plt_vma, rel):
	PLT_ENTRY_SIZE = 16
	return plt_vma + (i + 1) * PLT_ENTRY_SIZE
#end elf_x86_64_plt_sym_val(i, plt_vma, rel)

def elf_x86_plt_sym_val(i, plt_vma, rel):
	"""
	i386 and x86-64 are the same. Surprise.
	"""
	PLT_ENTRY_SIZE = 16
	return plt_vma + (i + 1) * PLT_ENTRY_SIZE
#end elf_x86_plt_sym_val(i, plt_vma, rel)


def bfd_elf_get_synthetic_symtab(sections, section_names, data, plt_section):
	"""
	A few simple things, most important is the parsing of .rela.plt.
	It allows us to get the correct function names for .plt.
	See rel.plt.py for the c code it's trying to replace.
	"""
	elfclass = sections[0].elfclass
	if '.rela.plt' in section_names:
		relplt_section_i = section_names.index('.rela.plt')
		if elfclass == elf1.ELFCLASS64:
			data_format = 'QQQ'
			data_len = 24
		else:
			data_format = 'III'
			data_len = 12
		#end if
	elif '.rel.plt' in section_names:
		relplt_section_i = section_names.index('.rel.plt')
		if elfclass == elf1.ELFCLASS64:
			data_format = 'QQ'
			data_len = 16
		else:
			data_format = 'II'
			data_len = 8
		#end if
	else:
		# No .rel.plt probably means it's statically linked maybe.
		return []
	#end if
	relplt = sections[relplt_section_i]
	relplt_data = data[relplt.sh_offset:relplt.sh_offset+relplt.sh_size]
	# rela.plt: offset    info         addend
	# rel.plt: offset info
	relplt_int = [x for x in struct.unpack('<' + (data_format * (len(relplt_data)//data_len)), relplt_data)]
	relpltx = [relplt_int[i*len(data_format):i*len(data_format)+len(data_format)] for i in range(len(relplt_int)//len(data_format))]
	#print("rpx", relplt_int, relpltx)
	plt_vma = plt_section.sh_addr
	
	# Not the same, but equivalent.
	#count = relplt.sh_size // relplt.sh_entsize
	count = len(relpltx)
	#print(count, len(relpltx))
	#size = count * sizeof (asymbol)
	ret = []
	p = 0 #relplt.relocation
	n = 0
	for i in range(count):
		# In this function we add plt_vma to (i+1)*16
		if elfclass == elf1.ELFCLASS64:
			addr = elf_x86_64_plt_sym_val(i, plt_vma, p)
		else:
			addr = elf_x86_plt_sym_val(i, plt_vma, p)
		#end if
		# Instead of giving them the string, give them the index to 
		# dynsym_names.
		if elfclass == elf1.ELFCLASS64:
			name_i = (relpltx[i][1]) >> 32
		else:
			name_i = (relpltx[i][1]) >> 8
		#end if
		ret.append((addr, name_i))
	#next i
	return ret
#end def bfd_elf_get_synthetic_symtab(sections, section_names, data, plt_section)

def demangle(s):
	"""
	An attempt to demangle. The C version of demangle in binutils is 5000 
	lines so I won't get a very good coverage in 100 lines.
	c++filt is a quick solution to this, but calling c++filt thousands of 
	times is very inefficient.
	_Z21LoadLocalAchievementsv
	_Z1fs
	_ZSt16__throw_bad_castv
	_ZNSs6appendEPKcm
	_ZNKSs16find_last_not_ofEPKcmm
	_ZSt18_Rb_tree_decrementPKSt18_Rb_tree_node_base
	FIXME: _ZNSt8ios_base4InitC1Ev
	P is pointer
	K is const
	c is char
	s is short
	m is unsigned long
	St is std::
	NSs is std::basic_string<char, std::char_traits<char>, std::allocator<char> >::
	E is currently unknown.
	"""
	if s[:2] != '_Z':
		return s
	#end if
	print(s)
	output = ''
	suffix = ''
	prefix = ''
	func_suffix = ''
	pos = 2
	prevVar = False
	hasFunc = False
	func_name = ''
	while pos < len(s):
		if s[pos:pos+1] in string.digits:
			# Function or named type
			lenstr = s[pos:pos+1]
			if s[pos+1:pos+2] in string.digits:
				lenstr = s[pos:pos+2]
				if s[pos+2:pos+3] in string.digits:
					lenstr = s[pos+1:pos+3]
				#end if
			#end if
			func_name_len = int(lenstr)
			pos += len(lenstr)
			func_name = s[pos:pos+func_name_len]
			output += prefix + func_name + suffix
			if not hasFunc:
				output += '('
				hasFunc = True
			#end if
			pos += func_name_len
			prevVar = False
		elif s[pos:pos+1] == 'v':
			# Void
			if prefix != '' or suffix != '':
				output += prefix + 'void' + suffix
			#end if
			pos += 1
		elif s[pos:pos+1] == 'P':
			# pointer
			suffix += '*'
			pos += 1
		elif s[pos:pos+1] == 'K':
			# const
			suffix = ' const' + suffix
			pos += 1
		elif s[pos:pos+1] == 'c':
			# char
			if prevVar: output += ', '
			output += prefix + 'char' + suffix
			prefix = ''
			suffix = ''
			prevVar = True
			pos += 1
		elif s[pos:pos+1] == 'm':
			# unsigned long
			if prevVar: output += ', '
			output += prefix + 'unsigned long' + suffix
			prefix = ''
			suffix = ''
			prevVar = True
			pos += 1
		elif s[pos:pos+1] == 's':
			# short
			if prevVar: output += ', '
			output += prefix + 'short' + suffix
			prefix = ''
			suffix = ''
			prevVar = True
			pos += 1
		elif s[pos:pos+1] == 'i':
			# int
			if prevVar: output += ', '
			output += prefix + 'int' + suffix
			prefix = ''
			suffix = ''
			prevVar = True
			pos += 1
		elif s[pos:pos+1] == 'b':
			# bool
			if prevVar: output += ', '
			output += prefix + 'bool' + suffix
			prefix = ''
			suffix = ''
			prevVar = True
			pos += 1
		elif s[pos:pos+2] == 'St':
			# std::
			output += 'std::'
			pos += 2
		elif s[pos:pos+3] == 'NSs':
			# string?
			output += 'std::basic_string<char, std::char_traits<char>, std::allocator<char> >::'
			func_name = 'basic_string'
			pos += 3
		elif s[pos:pos+4] == 'NKSs':
			# const type of above
			output += 'std::basic_string<char, std::char_traits<char>, std::allocator<char> >::'
			func_name = 'basic_string'
			func_suffix = ' const'
			pos += 4
		elif s[pos:pos+5] == 'NSols':
			output += 'std::basic_ostream<char, std::char_traits<char> >::operator<<('
			func_name = 'basic_ostream'
			pos += 5
		elif s[pos:pos+1] == 'E':
			# Unknown so far.
			pos += 1
		elif s[pos:pos+6] == 'RKSaIc':
			if prevVar: output += ', '
			output += 'std::allocator<char> const&'
			prefix = ''
			suffix = ''
			prevVar = True
			pos += 6
		elif s[pos:pos+3] == 'C1E':
			# Constructor
			output += func_name + '('
			pos += 3
		elif s[pos:pos+2] == 'dl':
			# delete
			output += 'operator delete('
			prevVar = False
			pos += 2
		elif s[pos:pos+2] == 'na':
			output += 'operator new[]('
			prevVar = False
			pos += 2
		elif s[pos:pos+2] == 'nw':
			output += 'operator new[]('
			prevVar = False
			pos += 2
		elif s[pos:pos+2] == 'ne':
			output += 'operator!=('
			prevVar = False
			pos += 2
		else:
			output += '?' + s[pos:pos+1] + '?'
			pos += 1
		#end if
	#loop
	return output + ')' + func_suffix
#end def demangle(s)

def find__x86_get_pc_thunk_bx(sect_name, sect_disasm, dynsym_addrs, dynsym_names):
	"""
	Solve x86 32-bit PIE using the function:
	__x86.get_pc_thunk.bx:
		520: MOV EBX, [ESP]
		523: RET
	"""
	# The easy way is to use the entry in symtab.
	if '__x86.get_pc_thunk.bx' in dynsym_names:
		raddr = dynsym_addrs[dynsym_names.index('__x86.get_pc_thunk.bx')]
		return raddr
	#end if
	if sect_disasm == None: return None
	# The hard way is to look for MOV EBX, [ESP]
	for i, instr in enumerate(sect_disasm):
		if instr[2] == 'MOV EBX, [ESP]':
			if (len(sect_disasm) > i + 1) and sect_disasm[i+1][2] == 'RET':
				print(instr)
				raddr = instr[0]
				return raddr
			else:
				# Very unlikely false positive?
				if verbose:
					print("Warning: __x86.get_pc_thunk.bx false positive.")
				#end if
			#end if
		#end if
	#next instr
	return None
#end def find__x86_get_pc_thunk_bx(sect_name, sect_disasm, dynsym_addrs, dynsym_names)

def disassemble(data, printFunction, printErrorFunction):
	"""
	data is a byte array with the ELF file.
	printFunction is a function that takes the following arguments:
	(sect_name, sect_disasm, dynsym_addrs, dynsym_names)
	printErrorFunction is a function that takes the following arguments:
	(*args, **kwargs)
	That's right, it should be a wrapper for print.
	"""
	(init_section, plt_section, text_section, fini_section, init_disasm, 
		plt_disasm, text_disasm, fini_disasm) = (None, None, None, 
		None, None, None, None, None)
	
	elf_st, sections, progs = elf1.parseELF(data)
	
	elfclass = elf_st.elfclass
	# TODO: Detect ARM, PPC, etc.
	DistormDecode_Const = distorm.Decode64Bits
	if elfclass == elf1.ELFCLASS32:
		DistormDecode_Const = distorm.Decode32Bits
	#end if
	
	section_names = [s.sh_name_str for s in sections]
	# If they don't have .text, we're in bad shape.
	if '.text' not in section_names:
		# TODO: use program headers.
		printError("Warning: .text is missing, using program headers.")
		execProgs = []
		for prog in progs:
			# 1 is exec, 4 is read, 2 is write
			if prog.p_flags & 0x1 == 1:
				execProgs.append(prog)
			#end if
		#next prog
		skipFirst = True
		for prog in execProgs:
			sec_name = '.text'
			if skipFirst:
				sec_name = '.init'
				skipFirst = False
			elif '.text' in section_names:
				sec_name = '.other_exec2'
			#end if
			section_names.append(sec_name)
			# Create a new section with all nulls
			add_section = elf1.section_header('', elf_st, len(sections), 'ignore')
			 # FIXME: Is this correct?
			add_section.sh_name_str = sec_name
			add_section.sh_type = 12345
			add_section.sh_flags = prog.p_flags
			add_section.sh_addr = prog.p_vaddr
			add_section.sh_offset = prog.p_offset
			add_section.sh_size = prog.p_filesz
			add_section.sh_link = 0
			add_section.sh_info = 0
			add_section.sh_addralign = 0
			add_section.sh_entsize = 0
			sections.append(add_section)
		#return None
		if '.text' not in section_names:
			section_names = [a == '.init' and '.text' or a for a in section_names]
		#end if
		if '.text' not in section_names:
			printError("Error: program headers lacks executable section")
			return None
		#end if
	#end if
	text_section_i = section_names.index('.text')
	text_section = sections[text_section_i]
	text_data = data[text_section.sh_offset:text_section.sh_offset+text_section.sh_size]
	
	dynamicallyLinked = False
	# If they don't have .plt, continue anyway.
	if '.plt' in section_names:
		dynamicallyLinked = True
		plt_section_i = section_names.index('.plt')
		plt_section = sections[plt_section_i]
		plt_data = data[plt_section.sh_offset:plt_section.sh_offset+plt_section.sh_size]
	#end if
	
	# If they don't have .dynstr, continue anyway.
	dynstr_data = None
	if '.dynstr' in section_names:
		dynstr_section = sections[section_names.index('.dynstr')]
		dynstr_data = data[dynstr_section.sh_offset:dynstr_section.sh_offset+dynstr_section.sh_size]
		if len(dynstr_data) == 0 or b'\x00' not in dynstr_data:
			printError("Error: .dynstr is not well-formed.", repr(dynstr_data))
			exit(1)
		#end if
		#print(repr(dynstr_data))
	#end if

	# If they don't have .dynsym, continue anyway.
	dynsym_data = None
	if '.dynstr' in section_names:
		dynsym_section = sections[section_names.index('.dynsym')]                         
		dynsym_data = data[dynsym_section.sh_offset:dynsym_section.sh_offset+dynsym_section.sh_size]  
	#end if
	
	sym_names = []
	sym_addrs = []
	# If they don't have .symtab, continue anyway.
	if '.strtab' in section_names and '.symtab' in section_names:
		strtab_section = sections[section_names.index('.strtab')]
		strtab_data = data[strtab_section.sh_offset:strtab_section.sh_offset+strtab_section.sh_size]
		if len(strtab_data) == 0 or b'\x00' not in strtab_data:
			printError("Error: .strtab is not well-formed.", repr(strtab_data))
			exit(1)
		#end if
		#print(repr(strtab_data))

		symtab_section = sections[section_names.index('.symtab')]                         
		symtab_data = data[symtab_section.sh_offset:symtab_section.sh_offset+symtab_section.sh_size]  
		
		# name info other shndx value size for 64-bit
		"""
		From man 5 elf:
		It requires two totally different code paths.
		typedef struct {
			uint32_t      st_name;
			unsigned char st_info;
			unsigned char st_other;
			uint16_t      st_shndx;
			Elf64_Addr    st_value;
			uint64_t      st_size;
		} Elf64_Sym;
		
		typedef struct {
			uint32_t      st_name;
			Elf32_Addr    st_value;
			uint32_t      st_size;
			unsigned char st_info;
			unsigned char st_other;
			uint16_t      st_shndx;
		} Elf32_Sym;
		
		"""
		if elfclass == elf1.ELFCLASS64:
			data_format = 'IBBHQQ'
			data_len = 24
		else:
			data_format = 'IIIBBH'
			data_len = 16
		#end if
		syms = [struct.unpack('<' + data_format, symtab_data[data_len*i:data_len*i+data_len]) for i in range(len(symtab_data)//data_len)] 
		try:
			sym_names = [strtab_data[x[0]:strtab_data.index('\x00', x[0])] for x in syms if x[1] != elf1.STT_SECTION or x[0] != 0]
		except ValueError:
			printError("Error: dynstr names were invalid.", [x[0] for x in syms])
			sym_names = ['' for x in syms]
		#end try
		# So we allow section types if they have names.
		# But we don't allow section types without names.
		# This will eventually need to be fine tuned for malicioius binaries.
		if elfclass == elf1.ELFCLASS64:
			sym_addrs = [x[4] for x in syms if x[1] != elf1.STT_SECTION or x[0] != 0]
		else:
			# 32-bit code path. Addrs are second.
			sym_addrs = [x[1] for x in syms if x[1] != elf1.STT_SECTION or x[0] != 0]
		#end if
		#print('\n'.join([str(x) for x in syms]))
		#print(sym_names, ', '.join([hex(x) for x in sym_addrs]))
		# main's address is in there twice. First with a null string.
	#end if

	# If they don't have .init, continue anyway.
	init_data = None
	if '.init' in section_names:
		init_section = sections[section_names.index('.init')]
		init_data = data[init_section.sh_offset:init_section.sh_offset+init_section.sh_size]
	#end if
	
	# If they don't have .fini, continue anyway.
	fini_data = None
	if '.fini' in section_names:
		fini_section = sections[section_names.index('.fini')]
		fini_data = data[fini_section.sh_offset:fini_section.sh_offset+fini_section.sh_size]
	#end if

	dynsym_names = []
	dynsym_addrs = []
	if dynsym_data and dynstr_data:
		if elfclass == elf1.ELFCLASS64:
			data_format = 'IBBHQQ'
			data_len = 24
		else:
			data_format = 'IIIBBH'
			data_len = 16
		#end if
		dynsyms = [struct.unpack('<' + data_format, dynsym_data[data_len*i:data_len*i+data_len]) for i in range(len(dynsym_data)//data_len)] 
		try:
			dynsym_names = [dynstr_data[x[0]:dynstr_data.index('\x00', x[0])] for x in dynsyms]
			#dynsym_names = [dynstr_data[x[0]:dynstr_data.index('\x00', x[0])] for x in dynsyms if x[1] != elf1.STT_SECTION or x[0] != 0]
		except ValueError:
			printError("Error: dynstr names were invalid.", [x[0] for x in dynsyms])
			dynsym_names = ['' for x in dynsyms]
		#end try
		# So we allow section types if they have names.
		# But we don't allow section types without names.
		# This will eventually need to be fine tuned for malicioius binaries.
		if elfclass == elf1.ELFCLASS64:
			dynsym_addrs = [x[4] for x in dynsyms if x[1] != elf1.STT_SECTION or x[0] != 0]
		else:
			# 32-bit code path. Addrs are second.
			dynsym_addrs = [x[1] for x in dynsyms if x[1] != elf1.STT_SECTION or x[0] != 0]
		#end if
	#end if

	"""
	dynsyms = [struct.unpack('<HHHHQHHHH', dynsym_data[24*i:24*i+24]) for i in range(len(dynsym_data)//24)] 
	try:
		dynsym_names = [dynstr_data[x[0]:dynstr_data.index('\x00', x[0])] for x in dynsyms]
	except ValueError:
		print("Error: dynstr names were invalid.", [x[0] for x in dynsyms])
		dynsym_names = ['' for x in dynsyms]
	#end try
	dynsym_addrs = [x[4] for x in dynsyms]
	print(dynsym_names, ', '.join([hex(x) for x in dynsym_addrs]))
	"""

	# Get the rest of the addresses
	
	if '.gnu.hash' in section_names:
		gnu_hash_section = sections[section_names.index('.gnu.hash')]                         
		gnu_hash_data = data[gnu_hash_section.sh_offset:gnu_hash_section.sh_offset+gnu_hash_section.sh_size]  
		(nbuckets, symndx, maskwords, shift2) = struct.unpack('<IIII', gnu_hash_data[:16])
		"""
		print "Buckets:", nbuckets
		print "symndx:", symndx
		print "maskwords:", maskwords
		print "shift2:", shift2
		bloom_filter = struct.unpack('<' + ('Q'*maskwords), gnu_hash_data[16:16+8*maskwords])
		print "Bloom filter:", bloom_filter
		pos = 16+8*maskwords
		third_section = struct.unpack('<' + ('I'*nbuckets), gnu_hash_data[pos:pos+nbuckets*4])
		for x in third_section:
			if x == 0: continue
			print dynsym_names[x], hex(dl_new_hash(dynsym_names[x]) % nbuckets)
		#next x
		pos += nbuckets*4
		dynsymcount = len(dynsyms)
		fourth_count = dynsymcount - symndx
		fourth_part = struct.unpack('<' + ('I'*fourth_count), gnu_hash_data[pos:pos+fourth_count*4])
		print "fourth_part:", fourth_part
		
		if pos+fourth_count*4 != len(gnu_hash_data):
			print "Error: did not consume all of .gnu.hash, probably due to an architecture issue."
			exit(1)
		#end if
		"""
	#end if
	#print('sa', sym_addrs, sym_names)
	if dynamicallyLinked:
		synthsym = bfd_elf_get_synthetic_symtab(sections, section_names, data, plt_section)
		for addr, dynsym_name_index in synthsym:
			if dynsym_name_index >= len(dynsym_addrs):
				printError("Error: synthsym has an too many symbols.", '%08x' % addr, dynsym_name_index, len(dynsym_addrs))
				continue
			#end if
			if dynsym_addrs[dynsym_name_index] != 0:
				# Validate
				if dynsym_addrs[dynsym_name_index] != addr:
					printError('Error: invalid .plt name compute', addr, '!=', dynsym_addrs[dynsym_name_index], dynsym_name_index, dynsym_names[dynsym_name_index])
				#end if
			dynsym_addrs[dynsym_name_index] = addr
		#next addr
		# Now that we have the correct addresses, add them to symbols.
		#print('da', dynsym_addrs, dynsym_names)
		sym_addrs.extend(dynsym_addrs)
		sym_names.extend(dynsym_names)
	#end if
	#print(synthsym)
	#print(zip(dynsym_names, dynsym_addrs))
	
	text_disasm = distorm.Decode(text_section.sh_addr, text_data, DistormDecode_Const)
	get_pc_thunk_bx = None
	if elfclass == elf1.ELFCLASS32:
		# We assume it's in text.
		get_pc_thunk_bx = find__x86_get_pc_thunk_bx(".text", text_disasm, sym_addrs, sym_names)
		if get_pc_thunk_bx: print("get_pc_thunk_bx:", hex(get_pc_thunk_bx))
	#end if
	if init_data:
		init_disasm = distorm.Decode(init_section.sh_addr, init_data, DistormDecode_Const)
		printFunction(".init", init_disasm, sym_addrs, sym_names, elf_st, sections, progs, data, get_pc_thunk_bx)
	#end if
	
	if dynamicallyLinked:
		plt_disasm = distorm.Decode(plt_section.sh_addr, plt_data, DistormDecode_Const)
		printFunction(".plt", plt_disasm, sym_addrs, sym_names, elf_st, sections, progs, data, get_pc_thunk_bx)
	#end if
	
	printFunction(".text", text_disasm, sym_addrs, sym_names, elf_st, sections, progs, data, get_pc_thunk_bx)

	if fini_data:
		fini_disasm = distorm.Decode(fini_section.sh_addr, fini_data, DistormDecode_Const)
		printFunction(".fini", fini_disasm, sym_addrs, sym_names, elf_st, sections, progs, data, get_pc_thunk_bx)
	#end if
	# A very handy set of information for writing your own disassembler.
	# TODO: make it possible to get these values from a library call.
	return (elf_st, sections, progs, section_names, 
		init_section, plt_section, text_section, fini_section, 
		sym_names, sym_addrs,
		init_disasm, plt_disasm, text_disasm, fini_disasm, get_pc_thunk_bx)
#end def main()

def deref(addr, data, elf_st, sections, progs, derefType='string'):
	for section in sections:
		#print('deref', addr, section.sh_addr, section.sh_addr + section.sh_size, section.sh_name_str)
		if addr >= section.sh_addr and \
			addr < (section.sh_addr + section.sh_size):
			# It's here.
			if derefType == 'string':
				section_data = data[section.sh_offset:section.sh_offset+section.sh_size]
				if b'\x00' not in section_data[addr-section.sh_addr:]:
					# Almost guaranteed this will never occur.
					return section_data[addr-section.sh_addr:]
				#end if
				#print("found string in section", section.sh_name_str, addr, hex(section.sh_addr), section.sh_size)
				end = section_data.index(b'\x00', addr-section.sh_addr)
				return section_data[addr-section.sh_addr:end]
			elif derefType == 'word':
				section_data = data[section.sh_offset:section.sh_offset+section.sh_size]
				pos = addr-section.sh_addr
				if len(section_data) <= pos+4:
					return None
				#end if
				#print('deref word', repr(section_data[pos:pos+4]))
				# TODO: endianness, signedness, and available data
				output = struct.unpack('<I', section_data[pos:pos+4])
				return output[0]
			#end if
		#end if
	#next section
	return None
#end def deref(addr, data, elf_st, sections, progs)

def cmpToC(cmpInstr, jmpStr, inverse = True):
	"""
	Example:
	cmpToC([0, '1234', 'CMP [RBP-0x4], 0x1'], 'JG 0x4005b4')
	TODO: this whole function is based on my guesses.
	FIXME: signed needs research.
	inverse is for while loops where logic is inverse.
	"""
	output = ''
	isCmp  = False
	isTest = False
	if cmpInstr[2].startswith('CMP '):
		isCmp = True
	else:
		isTest = True
	#end if
	# Comparator is the opposite of the jump.
	comparator = None
	if jmpStr.startswith('JG '):
		if inverse:
			comparator = '<='
		else:
			comparator = '>'
		#end if
	elif jmpStr.startswith('JGE '):
		if inverse:
			comparator = '<'
		else:
			comparator = '>='
		#end if
	elif jmpStr.startswith('JL '):
		if inverse:
			comparator = '>='
		else:
			comparator = '<'
		#end if
	elif jmpStr.startswith('JLE '):
		if inverse:
			comparator = '>'
		else:
			comparator = '<='
		#end if
	elif jmpStr.startswith('JA '):
		if inverse:
			comparator = '<='
		else:
			comparator = '>'
		#end if
	elif jmpStr.startswith('JAE '):
		if inverse:
			comparator = '<'
		else:
			comparator = '>='
		#end if
	elif jmpStr.startswith('JB '):
		if inverse:
			comparator = '>='
		else:
			comparator = '<'
		#end if
	elif jmpStr.startswith('JBE '):
		if inverse:
			comparator = '>'
		else:
			comparator = '<'
		#end if
	elif jmpStr.startswith('JZ '):
		if inverse:
			comparator = '!='
		else:
			comparator = '=='
		#end if
	elif jmpStr.startswith('JNZ '):
		if inverse:
			comparator = '=='
		else:
			comparator = '!='
		#end if
	elif jmpStr.startswith('JS '):
		if inverse:
			comparator = 'is not signed'
		else:
			comparator = 'is signed'
		#end if
	elif jmpStr.startswith('JNS '):
		if inverse:
			comparator = 'is signed'
		else:
			comparator = 'is not signed'
		#end if
	#end if
	v = cmpInstr[2].split(' ', 1)
	if len(v) != 2:
		print("Error: distorm failed to create a valid compare", cmpInstr[2])
		return ''
	#end if
	if ', ' not in v[1]:
		print("Error: distorm failed to create a valid compare", cmpInstr[2])
		return ''
	#end if
	# This may not work if distorm outputs weird data.
	x, y = v[1].split(', ', 1)
	
	# TODO: Validate all this.
	if isTest and x == y:
		if comparator == '!=':
			# TEST RAX, RAX; JZ 0x4004de; CALL RAX; ADD RSP, 0x8; 
			# RET
			return x
		elif comparator == '==':
			return '!' + x
		elif comparator == '>':
			return x + ' > 0'
		elif comparator == '>=':
			return x + ' >= 0'
		elif comparator == '<':
			return x + ' < 0'
		elif comparator == '<=':
			return x + ' <= 0'
		elif comparator == 'is not signed':
			return x + ' >= 0'
		elif comparator == 'is signed':
			return x + ' < 0'
		else:
			return "??" + x
		#end if
	#end if
	if comparator:
		output = x + ' ' + comparator + ' ' + y
	else:
		output = x + ' ??' + jmpStr + '= ' + y
	#end if
	return output
#end def cmpToC(cmpInstr, jmpStr)

def printDisassembly(sect_name, sect_disasm, dynsym_addrs, dynsym_names, elf_st, sections, progs, data, get_pc_thunk_bx):
	print(sect_name + ':')
	functions = []
	for instr in sect_disasm:
		if instr[2][:7] == 'CALL 0x':
			addr = int(instr[2][7:], 16)
			functions.append(addr)
		#end if
	#next instr
	endCurlyBraces = []
	prevTest = None
	prevCmp  = None
	prevEBX  = None
	flattenEBX = False
	currentFunction = 0
	arguments = []
	
	for i, instr in enumerate(sect_disasm):
		if flattenEBX:
			# This variable needs to be unset unless we set it.
			prevEBX = None
			flattenEBX = False
		else:
			# Flatten it the next time through.
			#flattenEBX = True
			pass
		#end if
		if instr[0] in dynsym_addrs:
			func_name = dynsym_names[dynsym_addrs.index(instr[0])]
			print(demangle(func_name) + ':')
			currentFunction = i
		elif instr[0] in functions:
			print('sub_%x:' % instr[0])
			currentFunction = i
		#end if
		if currentFunction == i and elf_st.elfclass == elf1.ELFCLASS64:
			q = findFirstReadReg(sect_name, sect_disasm, currentFunction, 'RDI|EDI')
			arguments = []
			firstArg = len(q) == 3 and q[0] != None
			if firstArg:
				print("Arguments:\n\t%s: int x = %s" % (q[0], q[1]))
				arguments.append((q[0], q[1], 'x'))
			#end if
			r = findFirstReadReg(sect_name, sect_disasm, currentFunction, 'RSI|ESI')
			if len(r) == 3 and r[0] != None:
				if not firstArg: print("Arguments:")
				firstArg = True
				print("\t%s: int y = %s" % (r[0], r[1]))
				arguments.append((r[0], r[1], 'y'))
			#end if
			r = findFirstReadReg(sect_name, sect_disasm, currentFunction, 'RDX|EDX')
			if len(r) == 3 and r[0] != None:
				if not firstArg: print("Arguments:")
				firstArg = True
				print("\t%s: int z = %s" % (r[0], r[1]))
				arguments.append((r[0], r[1], 'z'))
			#end if
			#print('arguments:', arguments)
		#end if
		if instr[0] in endCurlyBraces:
			print('; }')
		#end if
		print(('%8x' % instr[0]).rjust(8, ' ') + ':', end=' ')
		if instr[2][:7] == 'CALL 0x':
			func_name = None
			addr = int(instr[2][7:], 16)
			if addr in dynsym_addrs:
				func_name = dynsym_names[dynsym_addrs.index(addr)]
			#end if
			is_pc_thunk_bx = False
			add = ''
			if addr == get_pc_thunk_bx:
				# EBX is now at the next instruction.
				prevEBX = int(instr[0] + instr[1])
				flattenEBX = False
				add = ' ; EBX = ' + hex(prevEBX)
				is_pc_thunk_bx = True
			#end if
			if func_name:
				print("CALL", instr[2][7:], '<' + demangle(func_name) + '>' + add)
			elif is_pc_thunk_bx:
				print("CALL", instr[2][7:], '<__x86.get_pc_thunk.bx>' + add)
			else:
				print("CALL",  instr[2][7:], '<sub_' + instr[2][7:] + '>' + add)
			#end if
		elif instr[2][:4] == 'MOV ':
			# TODO: Same as LEA here for MOV EAX, [EBX-0x10]
			z = None
			if ', 0x' in instr[2]:
				addr = int(instr[2][instr[2].index(', 0x')+4:], 16)
				z = None
				if addr > 0 and addr < 256:
					z = chr(addr)
				elif addr == 0:
					z = None
				else:
					z = deref(addr, data, elf_st, sections, progs)
				#end if
				#if z == None:
				#end if
				if z != None:
					z = ' ; ' + repr(z)
				#end if
			#end if
			#print('ello there', repr(z))
			if z == None:
				z = InstrToC(instr[2], arguments)
				if z == None:
					z = ''
				else:
					python_object = z[1]
					z = ' ; ' + z[0]
				#end if
				# non-verbose throws away data.
				if not verbose: z = ''
			#end if
			if z:
				print(instr[2] + z)
			else:
				print(instr[2])
			#end if
		elif instr[2][:5] == 'TEST ':
			prevTest = instr
			prevCmp  = None
			print(instr[2], ';')
		elif instr[2][:4] == 'CMP ':
			prevCmp  = instr
			prevTest = None
			print(instr[2], ';')
		elif instr[2][:1] == 'J' and instr[2][:3] != 'JMP':
			isIf = True
			isLoop = False
			dsth = instr[2].split(' ', 1)[1]
			dsti = 0
			if dsth[:2] == '0x':
				dsti = int(dsth, 16)
				if dsti <= instr[0]:
					# It's a doggone loop!
					isLoop = True
					isIf = False
				#end if
			#end if
			if prevCmp or prevTest:
				prevX = prevCmp or prevTest
				print(instr[2], end=' ')
				if isIf:
					print('; if(' + cmpToC(prevX,  instr[2]) + ') {')
				else:
					print('; } while(' + cmpToC(prevX,  instr[2], False) + ')')
				#end if
			else:
				print(instr[2])
			#end if
			prevCmp  = None
			prevTest = None
			if isIf:
				endCurlyBraces.append(dsti)
			#end if
		elif instr[2][:6] == 'MOVZX ':
			# TODO: Copied from LEA
			# May not be correct. Need tests.
			# MOVZX ECX, WORD [RIP+0x2141df]
			if '[RIP+' not in instr[2]:
				if (not prevEBX) or ('[EBX' not in instr[2]):
					z = InstrToC(instr[2], arguments)
					if z == None:
						z = ''
					else:
						python_object = z[1]
						z = ' ; ' + z[0]
					#end if
					# non-verbose throws away data.
					if not verbose: z = ''
					print(instr[2] + z)
					continue
				#end if
			#end if
			print(instr[2], end=' ')
			if '[RIP+' in instr[2]:
				rpos = instr[2].index('[RIP+')
				addr = int(instr[2][rpos+5:instr[2].index(']', rpos+5)], 16) + instr[0] + instr[1]
			else:
				# NOTE: This assumes we didn't jump here with a different ebx.
				rpos = instr[2].index('[EBX')
				addr = int(instr[2][rpos+4:instr[2].index(']', rpos+4)], 16) + prevEBX
			#end if
			# TODO: find out what they want.
			z = deref(addr, data, elf_st, sections, progs, 'word')
			if z == None:
				z = ''
			else:
				z = hex(z)
			#end if
			print(' ; ' + hex(int(addr)) + ' ' + z)
		elif instr[2][:4] == 'LEA ':
			# LEA RAX, [RIP+0x204d6d]
			if '[RIP+' not in instr[2]:
				if (not prevEBX) or ('[EBX' not in instr[2]):
					print(instr[2])
					continue
				#end if
			#end if
			print(instr[2], end=' ')
			if '[RIP+' in instr[2]:
				rpos = instr[2].index('[RIP+')
				addr = int(instr[2][rpos+5:instr[2].index(']', rpos+5)], 16) + instr[0] + instr[1]
			else:
				# NOTE: This assumes we didn't jump here with a different ebx.
				rpos = instr[2].index('[EBX')
				addr = int(instr[2][rpos+4:instr[2].index(']', rpos+4)], 16) + prevEBX
			#end if
			z = deref(addr, data, elf_st, sections, progs)
			print(' ; ' + hex(int(addr)) + ' ' + repr(z))
		elif instr[2][:3] == 'RET':
			# Find the return value of the current function.
			cr = findConstantReturn(sect_name, sect_disasm, currentFunction)
			z = ''
			if len(cr) == 2 and cr[0] != None:
				z = ' ; return ' + hex(cr[0]) + ";"
			#end if
			print(instr[2] + z)
		else:
			z = InstrToC(instr[2], arguments)
			if z == None:
				z = ''
			else:
				python_object = z[1]
				z = ' ; ' + z[0]
				if prevEBX != None:
					newEBX = python_object.stepEBX(prevEBX)
					if newEBX != prevEBX:
						z += ' ; EBX = ' + hex(newEBX)
					#end if
					prevEBX = newEBX
					flattenEBX = False
				#end if
			#end if
			# non-verbose throws away data.
			if not verbose: z = ''
			print(instr[2] + z)
		#end if
	#next instr
#end def printDisassembly(sect_name, sect_disasm, dynsym_addrs, dynsym_names, data, get_pc_thunk_bx)

"""
From automatere automated dynamic reverse engineering system by Javantea
"""

# A dictionary of how many arguments each function has, file-based.
function_arguments = {}
# FIXME: This needs to be found in /usr/share or src dir, don't use the current directory.
x = sys.argv[0].split(os.sep)
x.pop()
javre_dir = os.sep.join(x)
if os.path.exists(javre_dir + '/func_args.txt'):
	fadata = open(javre_dir + '/func_args.txt','r').read().split('\n')
elif os.path.exists('func_args.txt'):
	fadata = open('func_args.txt','r').read().split('\n')
else:
	print("Warning: func_args.txt is missing.")
	fadata = []
#end if
for fa in fadata:
	if fa.startswith('#'): continue
	v = fa.split(' ', 1)
	if len(v) != 2: continue
	function_arguments[v[0]] = int(v[1])
#next fa
call_re = re.compile('CALLQ? *([^ ]*) ?<?([^@>]*@?[^>]*)>?')
instr_re = re.compile('([A-Z]*) *([^ ,]*), ?([^ ]*)')
instr2_re = re.compile('([A-Z]*) *([0-9a-fx-]*)\(([^ ,]*),([^ ,]*),([^ ,)]*)\),([^ ,]*)')
instr3_re = re.compile('([A-Z]*) *([^ ,]*),([0-9a-fx-]*)\(([^ ,]*),([^ ,]*),([^ ,)]*)\)')
instr1_re = re.compile('([A-Z]*) *([^ ]*)')
instr2r_re = re.compile(r'([A-Z]*) *([^ ,]*), ?([^ ]*) \[([^\]]*)\]')
# TODO: Make sure this fits all possible registers in x86-64 and i386.
reg_re = re.compile(r'^[A-Z]*[0-9]*$')
immediate_re = re.compile(r'^-?0x[0-9a-f]*$')

c_keyword = {'ADD':'+', 'SUB':'-', 'SUBL':'-', 'SUBQ':'-', 
	'MUL':'*', 'IMUL':'*', 
	'DIV':'/', 'IDIV':'/', 'XOR':'^', 
	'AND':'&', 'ANDL':'|', 'ANDQ':'|', 'ANDB':'|', 
	'OR':'|', 'ORL':'|', 'ORQ':'|', 
	'SHR':'>>', 'SHL':'<<', 
	'MOV':'', 'MOVL':'', 'MOVQ':'', 'MOVD':'', 'MOVB':'-', 'ADDQ':'+', 
	'ADDL':'+', 'ADDW':'+', 'ADDB':'+', 'SAR':'>>', 'SAL':'<<', 
	'PXOR':'^', 'PSUBB':'-', 'POR':'|', 'PAND':'|'}
# TODO: More python objects
py_keyword = {'ADD':AsmAddImmReg, 'SUB':AsmSubImmReg, 'MOV':AsmMovImmReg, 'MUL':AsmMulImmReg}
# Where each argument is put when a call is made.
arg_value = ['RDI', 'RSI', 'RDX', 'RCX', 'R8', 'R9']

def InstrToC(instr, arguments=None):
	"""
	Experimental Decompilation support.
	Inputs one tuple with (addr, length, instruction_mnemonic) in the 
	format that distorm gives.
	Outputs None or a tuple with (comment to append to asm, python_object)
	python_object allows a person to understand what is going on so that 
	they can reproduce the effect in python. For example:
	python_object = AsmNoop() allows a person to query what happens to EBX? 
	We can say that EBX is ebx.
	python_object.stepEBX(44) == 44
	"""
	arg_looks = []
	if arguments:
		arg_looks = [q[0] for q in arguments] + [q[1] for q in arguments]
	#end if
	output = ''
	python_object = AsmNoop()
	if instr.startswith('J'):
		# TODO: Jump depending on whatever compare was done last.
		output += ("goto ")
	elif instr.startswith('CALL'):
		# TODO: figure out all the arguments.
		func_name = call_re.match(instr).groups()
		if func_name == None:
			print("Warning: Call regex failed.")
			func_name = ('', '')
		#end if
		func_addr, func_name = func_name
		if '@plt' in func_name:
			func_name = func_name.replace('@plt', '')
		#end if
		num_args = 1
		if func_name in function_arguments:
			num_args = function_arguments[func_name]
		#end if
		if func_name == '':
			func_name = func_addr
			if reg_re.match(func_name):
				func_name = '*' + func_name
			#end if
		#end if
		output += (func_name + '(')
		# If we don't know where it's going to be don't try to grab it yet.
		if num_args > len(arg_value): num_args = len(arg_value)
		arg_names = []
		for arg_i in range(num_args):
			# TODO: Get the value here statically or dynamically.
			arg_i_name = (arg_value[arg_i])
			#arg_i_printable = getValue(arg_i_value)
			#output += ("Arg %i:" % arg_i) + repr(arg_i_printable) + '\n'
			arg_names.append(arg_i_name)
		#next arg_i
		output += ', '.join(arg_names) + ')'
	elif instr.startswith('RET'):
		output += ("Returning")
	elif instr.startswith('TEST'):
		# test eax, 0x3 is like if (eax & 0x3)
		output += ("test")
	elif instr.startswith('CMP'):
		output += ("compare")
		# TODO: make it very easy to know the result of this compare.
		lastCompare = instr
	elif instr.startswith('LEAVE'):
		output += ("leave")
	elif instr.startswith('LEA'):
		output += ("load effective address")
	elif instr.startswith('CMOV'):
		output += ("conditional mov")
	elif instr.startswith('PUSH'):
		# TODO: pusha, pushf, memory, etc
		op, value = instr, None
		x = instr1_re.match(instr).groups()
		if x:
			op, value = x
		#end if
		data = ''
		"""
		if value[:2] == '0x':
			data = value
		elif reg_re.match(value):
			srcr = value
			# TODO: Get the value here statically or dynamically.
			reg_value = srcr
			# TODO: data
			# TODO: memory
			#if srcRegRef:
			data = value + ' = ' + srcValue
		else:
			# TODO: 
			# pushq 0x202ec2(%rip)
			data = value
		#end if"""
		# At this point, this is redundant.
		#output += (op + ' ' + value) # + '; ' +  data)
	elif instr.startswith('POP'):
		# TODO: popa, pushf, memory, etc
		op, value = instr, None
		x = instr1_re.match(instr).groups()
		if x:
			op, value = x
			# stuff here?
			output += op + ' ' + value
		else:
			output += (op)
		#end if
	elif instr.startswith('MOVZBL'):
		# There's a string here...
		arg = ''
		output += ("movzbl " + arg)
	elif instr.startswith('MOVZWL'):
		# There's a string here...
		arg = ''
		output += ("movzbl " + arg)
	elif instr.startswith('MOVSBQ'):
		# What is this?
		arg = ''
		output += ("movsbq " + arg)
	elif instr.startswith('MOVSLQ'):
		# What is this?
		arg = ''
		output += ("movslq " + arg)
	elif instr.startswith('MOVLPD'):
		# What is this?
		arg = ''
		output += ("movlpd " + arg)
	elif instr.startswith('MOVHPD'):
		# What is this?
		arg = ''
		output += ("movhpd " + arg)
	elif instr.startswith('MOVDQ'):
		# Move Double Quad aligned or unaligned
		arg = ''
		output += ("movdq " + arg)
	elif instr.startswith('VMOVDQ'):
		# Not entirely sure, but similar to movdq.
		arg = ''
		output += ("vmovdq " + arg)
	elif instr.startswith('MOVABS'):
		# Move Absolute value
		arg = ''
		output += ("movabs " + arg)
	elif instr.startswith('MOVZX'):
		# Move with Zero-Extend
		r = instr2r_re.match(instr)
		arg = ''
		if r:
			(op, dst, srcType, src) = r.groups()
			# TODO: srcType is either BYTE or WORD or ...
			if '+' in src:
				# Offset can be reg*imm
				srcreg, offset = src.split('+', 1)
				arg = dst + ' = ' + srcreg + '[' + offset + ']'
			else:
				arg = dst + ' = ' + src + '[0]'
			#end if
		else:
			r = instr_re.match(instr)
			if r:
				(op, dst, src) = r.groups()
				arg = dst + ' = ' + src
			#end if
		#end if
		# char EAX = RAX[0];
		output += (arg)
	elif instr.startswith('MOVSX'):
		# Move with Sign-Extension
		arg = ''
		output += ("movsx " + arg)
	elif instr.startswith('BNDMOV'):
		# What is this?
		arg = ''
		output += ("bndmov " + arg)
	elif instr.startswith('BT'):
		output += ("bit test")
	elif instr.startswith('PUNPCK'):
		# MMX Unpack low data or high data.
		output += ("MMX unpack")
	elif instr.startswith('PSHUF'):
		# MMX Shuffle packed low words, high words, or double words.
		output += ("MMX shuffle")
	elif instr.startswith('PMOV'):
		# MMX Move packed
		output += ("MMX move")
	elif instr.startswith('PCMP'):
		# MMX Compare packed
		output += ("MMX compare")
	elif instr.startswith('PSL') or instr.startswith('PSR'):
		# MMX Shift Left Logical many variations
		# ps conflicts with pshuff, but so long as this is after...
		output += ("MMX Shift")
	elif instr.startswith('BS'):
		# Bit scan forward or reverse
		output += ("Bit scan")
	elif instr.startswith('SBB'):
		output += ("Subtract with borrow")
	elif instr.startswith('XCHG'):
		# a = b, b = a without the whole business of a temporary variable.
		output += ("Exchange")
	elif instr.startswith('RO'):
		# Rotate left or right
		output += ("Rotate")
	elif instr.startswith('NOP'):
		#print("nop")
		# No need.
		return None
	else:
		dstNotReg = False
		srcNotReg = False
		x = instr_re.match(instr)
		if x:
			op, dst, src = x.groups()
			keyword = 'unknown'
			if op == 'XOR' and src == dst:
				# xor rax, rax is rax = 0.
				return (src + ' = 0', AsmMovImmReg(src, 0))
			#end if
			if not op in c_keyword:
				if verbose:
					print("Interesting op:", op)
				#end if
			else:
				keyword = c_keyword[op]
			#end if
			# TODO: Intel syntax immediate and ptr immediate conversion to C
			if ',' in dst:
				# FIXME: Intel syntax
				xsrc = None
				xdst = instr3_re.match(instr)
				if not xdst: xsrc = instr2_re.match(instr)
				if xdst:
					op, src, dst1,dst2,dst3,dst4 = xdst.groups()
					# Do the actual c conversion here:
					dst1op = dst1 and (dst1 + ' + ') or ''
					dst = dst1op + dst2 + ' + (' + dst3 + ' * ' + dst4 + ')'
					dstNotReg = True
				elif xsrc:
					op, src1, src2, src3, src4, dst = xsrc.groups()
					# Do the actual c conversion here:
					src1op = src1 and (src1 + ' + ') or ''
					src = src1op + src2 + ' + (' + src3 + ' * ' + src4 + ')'
					srcNotReg = True
				#end if
			#end if
			#end if
			
			#print('InstrToC src', src, 'dst', dst)
			if src in arg_looks:
				argument_i = arg_looks.index(src)
				if argument_i >= len(arguments): argument_i -= len(arguments)
				src = arguments[argument_i][2]
			#end if
			dstReg = False
			dstRegRef = False
			srcReg = False
			srcRegRef = False
			if not dstNotReg:
				if reg_re.match(dst):
					dstReg = True
				elif dst[:1] == '[':
					dstRegRef = True
				#end if
			#end if
			if not srcNotReg:
				if reg_re.match(src):
					srcReg = True
				elif src[:1] == '[':
					srcRegRef = True
				#end if
			#end if
			if dstReg or dstRegRef:
				regNext = dst
				derefNext = dstRegRef
			#end if
			#end = '\n'
			end = ' '
			if srcReg or srcRegRef: end = ' '
			output += (dst + " " + keyword + "= " + src + ';' + end)
			"""if srcReg or srcRegRef:
				srcr = src
				if ':' in src: srcr = src[:src.index(':')]
				# TODO: Get the value here statically or dynamically.
				reg_value = gdb.parse_and_eval(srcr.replace('%','$'))
				#print('type:', reg_value.type.name)
				if reg_value.type.name == 'int64_t':
					srcValue = hex(int(long(reg_value)))
				elif reg_value.type.name == 'vec128':
					# This can be many things, but we will be told which it is.
					srcValue = str(reg_value['v4_float'][0])
				elif reg_value.type.name == 'void':
					srcValue = '(void)'
				elif reg_value.type.name == None:
					srcValue = hex(int(long(reg_value)))
				else:
					srcValue = '(' + reg_value.type.name + ')' + hex(int(long(reg_value)))
				#end if
				# TODO: data
				# TODO: memory
				#if srcRegRef:
				#output += (srcr + ' = ' + srcValue)
				#regNext = None
			#end if"""
			if dstReg and op in py_keyword:
				srcImm = immediate_re.match(src)
				if srcImm:
					python_object = py_keyword[op](dst, int(src, 16))
				#end if
			#else:
			#	print("Interesting op:", op)
			#end if
		#end if
	#end if
	return output, python_object
#end def InstrToC(instr, arguments=None)

"""
End of Automatere code.
"""

mov_eax_imm_re = re.compile(r'^MOV (RAX|EAX|AX|AL|AH), (0x[0-9a-f]*)$')
mov_eax_re = re.compile(r'^MOV (RAX|EAX|AX|AL|AH), ')
modify_eax_re = re.compile(r'^[^ ]* (RAX|EAX|AX|AL|AH), ')

def findConstantReturn(sect_name, sect_disasm, function_index):
	"""
	Given 
	f0:
		645: PUSH RBP ; 
		646: MOV RBP, RSP
		649: MOV EAX, 0x17 ; '\x17'
		64e: POP RBP ; POP RBP
		64f: RET ; Returning
	Find the value of EAX.
	Returns the value of the EAX register and the register name.
	For example, this code would result in the tuple (23, 'EAX') being 
	returned.
	We also have to make sure RAX, EAX, AX, AL, and AH are not modified.
	Easy enough...
	"""
	#print("findConstantReturn(", sect_name, sect_disasm[function_index:function_index+3], function_index, ")")
	eax_value = None
	register = None
	# Look for MOV EAX, immediate
	for i, instr in enumerate(sect_disasm[function_index:]):
		x = mov_eax_imm_re.match(instr[2])
		if x:
			m = x.groups()
			eax_value = int(m[1], 16)
			register = m[0]
			# Don't check whether they modify or ret.
			continue
		#end if
		if eax_value != None:
			# Check whether they modify eax or call a function.
			x = modify_eax_re.match(instr[2])
			if x:
				# Compare and test are reads, not writes.
				if instr[2].startswith('CMP') or instr[2].startswith('TEST'):
					continue
				#end if
				#print('match modify', instr[2])
				eax_value = None
				register = None
			#end if
			x = call_re.match(instr[2])
			if x:
				#print('match call', instr[2])
				eax_value = None
				register = None
			#end if
		#end if
		if instr[2].startswith('RET'):
			# TODO: We may or may not be done.
			break
		#end if
	#next instr
	return (eax_value, register)
#end def findConstantReturn(sect_name, sect_disasm, function_index)

def findFirstReadReg(sect_name, sect_disasm, function_index, reg_name):
	"""
	Given
	f1:
		645: PUSH RBP ; 
		646: MOV RBP, RSP
		649: MOV [RBP-0x4], EDI
		64c: MOV EAX, [RBP-0x4]
		64f: ADD EAX, 0x17 ; EAX += 0x17; 
		652: POP RBP ; POP RBP
		653: RET
	and EDI.
	Find the first read of EDI.
	Returns the destination that EDI is written to, the register name, and the address.
	For example, this code would result in the tuple ('[RBP-0x4]', 'EDI', 0x649)
	"""
	read_reg_re = re.compile(r'.*, (' + reg_name + ')$')
	modify_reg_re = re.compile(r'^[^ ]* (' + reg_name + '), ')
	#print('findFirstReadReg', reg_name)
	for i, instr in enumerate(sect_disasm[function_index:]):
		x = read_reg_re.match(instr[2])
		if x:
			#print('match', instr[2])
			m = x.groups()
			register = m[0]
			read_dst = instr[2][:-len(register)-2].split(' ', 1)
			if len(read_dst) == 2:
				return (register, read_dst[1], instr[0])
			#end if
			return (None, None, None)
		#end if
		x = modify_reg_re.match(instr[2])
		if x:
			if instr[2].startswith('CMP') or instr[2].startswith('TEST'):
				m = x.groups()
				register = m[0]
				# TODO: Think about this more.
				return (register, '<no reg>', instr[0])
			#end if
			break
		#end if
		if instr[2].startswith('RET'):
			# TODO: We may or may not be done.
			break
		#end if
	#next i, instr
	return (None, None, None)
#end def findFirstReadReg(sect_name, sect_disasm, function_index, reg_name)

def printError(*args, **kwargs):
	"""
	The correct way to wrap print.
	"""
	print(*args, **kwargs)
#end def def printError(*args, **kwargs)

def main():
	global verbose

	import argparse
	parser = argparse.ArgumentParser(description='Disassemble an ELF file.')
	parser.add_argument('filename', metavar='filename', nargs='?', type=str,
                   help='an ELF file', default='/bin/ls')
	parser.add_argument('--verbose', '-v', action='count')
	args = parser.parse_args(sys.argv[1:])
	filename = args.filename
	verbose = args.verbose
	data = open(filename,'rb').read()
	disassemble(data, printDisassembly, printError)
#end def main()

if __name__ == '__main__':
	main()
#end if
